#!/usr/bin/python3
# -*- coding: utf-8 -*-
##
## @author Edouard DUPIN
##
## @copyright 2023, Edouard DUPIN, all right reserved
##
## @license MPL v2.0 (see license file)
##
from .connection import (
    KaranageConnection,
    KaranageConnectionInterface,
    KaranageMock,
    KaranageResponse,
)
from .exception import KaranageException
from .log import GroupLogElement, KaranageLog
from .state import KaranageState, StateSystem



#!/usr/bin/python3
# -*- coding: utf-8 -*-
##
## @author Edouard DUPIN
##
## @copyright 2023, Edouard DUPIN, all right reserved
##
## @license MPL v2.0 (see license file)
##
from abc import ABC, abstractmethod
import copy
import enum
import json
from datetime import datetime, timedelta
from pathlib import Path
from typing import (
    Any,
    Dict,
    List,
    NamedTuple,
    Optional,
)
import time
import requests
from .exception import KaranageException


class KaranageResponse:
    def __init__(
        self, url: str, status: Optional[int] = None, data: Optional[str] = None
    ):
        self.url = url
        self.status = status
        self.data = data


class KaranageConnectionInterface(ABC):
    """Generic connection interface to access to karanage interface.

    :param ABC: Abstraction model class.
    """

    @abstractmethod
    def post(
        self,
        service: str,
        url_offset: Optional[str],
        data: Any = None,
        headers: Optional[Dict] = None,
        params: Optional[Dict] = None,
    ) -> KaranageResponse:
        """POST request on the server.

        :param service: Service on which we need to access.
        :param url_offset: Offset of the url server.
        :param data: Json data to post, defaults to None.
        :param headers: Headers value to add on the request, defaults to None.
        :param params: Parameters to send on the server, defaults to None.
        :return: Result of the request.
        """

    @abstractmethod
    def get(
        self,
        service: str,
        url_offset: Optional[str],
        headers: Optional[Dict] = None,
        params: Optional[Dict] = None,
    ) -> KaranageResponse:
        """GET request on the server.

        :param service: Service on which we need to access.
        :param url_offset: Offset of the url server.
        :param headers: Headers value to add on the request, defaults to None.
        :param params: Parameters to send on the server, defaults to None.
        :return: Result of the request.
        """


class KaranageConnection(KaranageConnectionInterface):
    """Connection on the karanage REST server.

    :param KaranageConnectionInterface: Model of connection.
    """

    def __init__(
        self,
        url: Optional[str] = None,
        group: Optional[str] = None,
        token: Optional[str] = None,
        config_file: Optional[str] = None,
        default_values: Optional[Dict[str, str]] = None,
    ) -> None:
        """Initialize the communication class.
        :param url: URL of the karanage API server.
        :param group: Group of the message (token need to have the autorotation to published on it).
        :param token: Token to validate the access on the application.
        :param config_file: path to the configuration file if overload.
        :param default_values: Default configurations.
        """
        self.url = "http://localhost:20080/karanage/api"
        self.group = "test"
        self.token = None
        # load user default value:
        if default_values is not None:
            if "url" in default_values:
                self.url = default_values["url"]
            if "group" in default_values:
                self.group = default_values["group"]
            if "token" in default_values:
                self.token = default_values["token"]
        # keep correct config file:
        if config_file is None:
            config_file = "/etc/karanage/connection.json"
        # check if the config exist:
        if Path(config_file).exists():
            f = open(config_file, "r")
            configuration = json.loads(f.read())
            f.close()
        else:
            configuration = {}
        # Update data with config file:
        if "url" in configuration:
            self.url = configuration["url"]
        if "group" in configuration:
            self.group = configuration["group"]
        if "token" in configuration:
            self.token = configuration["token"]
        # set user command - line configuration:
        if url is not None:
            self.url = url
        if group is not None:
            self.group = group
        if token is not None:
            self.token = token

    def get_url(self, service: str, url_offset: Optional[str] = None):
        """Get the formatted URL for specific service and specific offset

        :param service: Name of the service we want to connect on.
        :param url_offset: additional url path.
        :return: the specific url string (all before ?).
        """
        if url_offset is None:
            return f"{self.url}/{service}/{self.group}"
        return f"{self.url}/{service}/{self.group}/{url_offset}"

    def update_header(self, headers):
        """Add security header on the client header config.

        :param headers: _description_
        :return: _description_
        """
        if headers is None:
            headers = {}
        else:
            headers = copy.deepcopy(headers)
        if self.token is not None and len(self.token) > 15:
            headers["Authorization"] = f"zota {self.token}"
        return headers

    # @override
    def post(
        self,
        service: str,
        url_offset: Optional[str] = None,
        data: Any = None,
        headers: Optional[Dict] = None,
        params: Optional[Dict] = None,
    ) -> KaranageResponse:
        headers = self.update_header(headers)
        url = self.get_url(service, url_offset)
        ret = requests.post(url, json=data, headers=headers, params=params)
        try:
            return KaranageResponse(url, ret.status_code, ret.content.decode("utf-8"))
        except requests.exceptions.ConnectionError as ex:
            raise KaranageException(f"Fail connect server: '{url}'", 0, str(ex))

    # @override
    def get(
        self,
        service: str,
        url_offset: Optional[str] = None,
        headers: Optional[Dict] = None,
        params: Optional[Dict] = None,
    ) -> KaranageResponse:
        headers = self.update_header(headers)
        url = self.get_url(service, url_offset)
        ret = requests.get(url, headers=headers, params=params)
        try:
            return KaranageResponse(url, ret.status_code, ret.content.decode("utf-8"))
        except requests.exceptions.ConnectionError as ex:
            raise KaranageException(f"Fail connect server: '{url}'", 0, str(ex))


class MockData(NamedTuple):
    """Received Mock data from the karanage backend.

    :param NamedTuple: Tuple generic interface
    """
    type: str
    service: str
    url_offset: Optional[str]
    data: Any
    headers: Optional[Dict[str, Any]]
    params: Optional[Dict[str, Any]]


class KaranageMock(KaranageConnectionInterface):
    """Simple Karanage Mock to permit to abstract
    Note: This class is for test only !!!

    :param KaranageConnectionInterface: Model of connection.
    """

    def __init__(
        self,
    ) -> None:
        """Initialize the communication class."""
        self.request: List[MockData] = []

    def post(
        self,
        service: str,
        url_offset: Optional[str],
        data: Any = None,
        headers: Optional[Dict[str, Any]] = None,
        params: Optional[Dict[str, Any]] = None,
    ) -> KaranageResponse:
        self.request.append(
            MockData("POST", service, url_offset, data, headers, params)
        )
        return KaranageResponse(f"{service}/{url_offset}", 200, "{}")

    def get(
        self,
        service: str,
        url_offset: Optional[str],
        headers: Optional[Dict[str, Any]] = None,
        params: Optional[Dict[str, Any]] = None,
    ) -> KaranageResponse:
        self.request.append(MockData("GET", service, url_offset, None, headers, params))
        return KaranageResponse(f"{service}/{url_offset}", 200, "{}")

    def get_values(self, time_out: Optional[float] = None) -> List[MockData]:
        """get the list of last added values

        :param time_out: Timeout before exiting in error
        :returns: all collected values.
        """
        if time_out is not None and time_out > 0:
            start_time = datetime.now()
            while len(self.request) == 0:
                time.sleep(0.1)
                now = datetime.now()
                if now - start_time > timedelta(seconds=time_out):
                    return []
        out = self.request
        self.request = []
        return out


    def clear_values(self) -> None:
        """Clear all the received data."""
        self.request = []



#!/usr/bin/python3
# -*- coding: utf-8 -*-
##
## @author Edouard DUPIN
##
## @copyright 2023, Edouard DUPIN, all right reserved
##
## @license MPL v2.0 (see license file)
##
from datetime import datetime, timezone
import enum
import json
from typing import (
    Any,
    Dict,
    List,
    Optional,
)

import requests

from .connection import KaranageConnectionInterface
from .exception import KaranageException


class GroupLogElement:
    def __init__(self, id: int, data: str, time: Optional[datetime] = None):
        self.id = id
        self.time = time
        self.data = data


## Generic karanage sending system.
class KaranageLog:
    def __init__(
        self, connection: KaranageConnectionInterface, system: Optional[str] = None
    ) -> None:
        """Initialize the communication class.
        :param connection: Connection interface.
        """
        self.connection = connection
        self.system = system
        self.service = "log"

    def send(
        self,
        data: Dict,
        id: Optional[int] = None,
        uuid_group: Optional[int] = None,
        time: Optional[datetime] = None,
    ) -> None:
        """Send a message to the server.
        :param data: Data to send to the server
        :param id: Local internal ID
        :param uuid_group: local internal group UUID
        :param time_log: Receive time of the log
        """
        param: Dict[str, Any] = {}
        if id is not None:
            param["id"] = id
        if uuid_group is not None:
            param["uuid"] = uuid_group
        if time is not None:
            param["time"] = time.astimezone(timezone.utc).isoformat()
        ret = self.connection.post(self.service, self.system, data=data, params=param)
        if ret is None:
            raise KaranageException(
                f"Fail send message sub-library return None")
        if ret.status is None or not 200 <= ret.status <= 299:
            raise KaranageException(
                f"Fail send message: '{ret.url}'", ret.status, ret.data
            )

    def send_multiple(
        self, data_input: List[GroupLogElement], uuid_group: Optional[int] = None
    ) -> None:
        """Send multiple log message to the server.
        :param data: Data to send to the server
        :param uuid_group: local internal group UUID
        """
        data = []
        for elem in data_input:
            if elem.time is not None:
                data.append(
                    {
                        "id": elem.id,
                        "time": elem.time.astimezone(timezone.utc).isoformat(),
                        "data": elem.data,
                    }
                )
            else:
                data.append(
                    {
                        "id": elem.id,
                        "time": datetime.now().astimezone(timezone.utc).isoformat(),
                        "data": elem.data,
                    }
                )


        param = {}
        if uuid_group is not None:
            param["uuid"] = uuid_group
        ret = self.connection.post(
            self.service, f"{self.system}/push_multiple", data=data, params=param
        )
        if ret is None:
            raise KaranageException(
                f"Fail send message sub-library return None")
        if ret.status is None or not 200 <= ret.status <= 299:
            raise KaranageException(
                f"Fail send message: '{ret.url}'", ret.status, ret.data
            )



#!/usr/bin/python3
# -*- coding: utf-8 -*-
##
## @author Edouard DUPIN
##
## @copyright 2023, Edouard DUPIN, all right reserved
##
## @license MPL v2.0 (see license file)
##
import enum
import json
from typing import Any, Dict, Optional, Union, List, NamedTuple

from datetime import datetime, timezone
import requests

from .connection import KaranageConnectionInterface
from .exception import KaranageException


class StateSystem(enum.Enum):
    OK = "OK"
    FAIL = "FAIL"
    DOWN = "DOWN"

class StateData(NamedTuple):
    topic: str
    time: datetime
    state: str
    data: Dict

def convert_to_state_data(json_data: Dict, topic:Optional[str] = None) -> StateData:
    time = datetime.fromtimestamp(json_data["time"], timezone.utc)
    if topic is not None:
        topic = topic
    else:
        topic = json_data["topic"]
    return StateData(topic, time, json_data["state"], json_data["data"])



## Generic karanage sending system.
class KaranageState:
    def __init__(self, connection: KaranageConnectionInterface) -> None:
        """Initialize the communication class.
        :param connection: Connection interface.
        """
        self.connection = connection

    def send(
        self,
        topic: str,
        data: Optional[Dict] = None,
        state: StateSystem = StateSystem.OK,
    ) -> None:
        """Send a message to the server.
        :param topic: Topic where to publish the data.
        :param data: Data to send to the server
        :param state: State of the current system
        """
        if data is None:
            data = {}
        param = {}
        if state is not None:
            param["state"] = state.value
        ret = self.connection.post("state", topic, data=data, params=param)
        if ret is None:
            raise KaranageException(
                f"Fail send message sub-library return None")
        if ret.status is None or not 200 <= ret.status <= 299:
            raise KaranageException(
                f"Fail send message: '{ret.url}'", ret.status, ret.data
            )

    def get(self, topic: str, since: Union[None, str, datetime] = None) -> Optional[StateData]:
        """Get all the topic fom the server.
        :param since: ISO1866 time value.
        :return: A dictionary with the requested data.
        """
        param = {
            "mode":"raw" # request raw mode to have the timestamp in a float in second
        }
        if since is not None:
            if type(since) is str:
                param["since"] = since
            elif type(since) is datetime:
                param["since"] = since.replace(tzinfo=timezone.utc).isoformat()
            else:
                raise KaranageException(
                    f"Wrong input parameter type Must be a str or datetime: '{type(since)}'")
        ret = self.connection.get("state", topic, params=param)
        if ret is None:
            raise KaranageException(
                f"Fail send message sub-library return None")
        if ret.status is not None and 200 <= ret.status <= 299:
            if ret.data == "":
                return None
            return convert_to_state_data(json.loads(ret.data), topic)
        return None
        #raise KaranageException(f"Fail get data: '{ret.url}'", ret.status, ret.data)

    def gets(self, since: Union[None, str, datetime] = None) -> List[StateData]:
        """Get all the topic fom the server.
        :param since: ISO1866 time value.
        :return: A dictionary with the requested data.
        """
        param = {
            "mode":"raw" # request raw mode to have the timestamp in a float in second
        }
        if since is not None:
            if type(since) is str:
                param["since"] = since
            elif type(since) is datetime:
                param["since"] = since.replace(tzinfo=timezone.utc).isoformat()
            else:
                raise KaranageException(
                    f"Wrong input parameter type Must be a str or datetime: '{type(since)}'")
        ret = self.connection.get("state", params=param)
        if ret is None:
            raise KaranageException(
                f"Fail send message sub-library return None")
        if ret.status is not None and 200 <= ret.status <= 299:
            out: List[StateData] = []
            if ret.data is not None:
                for elem in json.loads(ret.data):
                    out.append(convert_to_state_data(elem))
            return out
        raise KaranageException(f"Fail get data: '{ret.url}'", ret.status, ret.data)

    def get_history(
        self,
        topic: Optional[str] = None,
        since: Union[None, str, datetime] = None,
        since_id: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> List[StateData]:
        """Get all the topic fom the server.
        :param since: ISO1866 time value.
        :param since_id: remote BDD index of the field.
        :param limit: the number of value we want to get
        :return: A dictionary with the requested data.
        """
        param:Dict[str, Any] = {

            "mode":"raw" # request raw mode to have the timestant in a float in second
        }
        if since is not None:
            if type(since) is str:
                param["since"] = since
            elif type(since) is datetime:
                param["since"] = since.replace(tzinfo=timezone.utc).isoformat()
            else:
                raise KaranageException(
                    f"Wrong input parameter type Must be a str or datetime: '{type(since)}' {type(str)}")
        if since_id is not None:
            param["sinceId"] = since_id
        if limit is not None:
            param["limit"] = limit
        ret = self.connection.get("state_history", topic, params=param)
        if ret is None:
            raise KaranageException(
                f"Fail send message sub-library return None")
        if ret.status is not None and 200 <= ret.status <= 299:
            out: List[StateData] = []
            if ret.data is not None:
                for elem in json.loads(ret.data):
                    out.append(convert_to_state_data(elem, topic))
            return out
        raise KaranageException(f"Fail get data: '{ret.url}'", ret.status, ret.data)



#!/usr/bin/python3
# -*- coding: utf-8 -*-
##
## @author Edouard DUPIN
##
## @copyright 2023, Edouard DUPIN, all right reserved
##
## @license MPL v2.0 (see license file)
##

from setuptools import setup
import os


def readme():
    with open("README.md") as f:
        return f.read()


def read_version_file():
    if not os.path.isfile("version.txt"):
        return ""
    file = open("version.txt", "r")
    data_file = file.read()
    file.close()
    if len(data_file) > 4 and data_file[-4:] == "-dev":
        data_file = data_file[:-4]
    return data_file


# https://pypi.python.org/pypi?%3Aaction=list_classifiers
setup(
    name="karanage",
    version=read_version_file(),
    description="Karanage interface is a simple interface to access to the karanage service to send and receive data",
    long_description=readme(),
    url="https://gitea.atria-soft.org/kangaroo-and-rabbit/karanage",
    author="Edouard DUPIN",
    author_email="yui.heero@gmail.com",
    license="MPL-2",
    packages=["karanage"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: Mozilla Public License 2.0 (MPL 2.0)",
        "Programming Language :: Python",
        "Operating System :: POSIX",
        "Topic :: Software Development :: Libraries",
    ],
    package_data={"karanage": ["py.typed"]},
    install_requires=["requests"],
    long_description_content_type="text/markdown",
    keywords="kar karanage",
    include_package_data=True,
    zip_safe=False,
)

# To develop: sudo ./setup.py install --user
#             sudo ./setup.py develop --user

# To register:
# https://packaging.python.org/en/latest/tutorials/packaging-projects/
# python3 -m build
# python3 -m twine upload dist/*



#!/usr/bin/python3
# -*- coding: utf-8 -*-
##
## @author Edouard DUPIN
##
## @copyright 2023, Edouard DUPIN, all right reserved
##
## @license MPL v2.0 (see license file)
##
from typing import Optional

class KaranageException(Exception):
    def __init__(self, message:str, error_id: Optional[int] = None, error_message: Optional[str] = None):
        # Call the base class constructor with the parameters it needs
        super().__init__(message)

        # Now for your custom code...
        self.error_id = error_id
        self.error_message = error_message

    def __str__(self):
        return f"{Exception.__str__(self)} status={self.error_id} message='{self.error_message}'"
